﻿
/*******************************************
 * 信号処理クラス
 * 
 * 更新履歴
 *  2010/5/23   森下整備開始
 *  2011/2/18   名前空間を名づけて、クラスっぽく改善
 *              渡されたデータ幅がFFTデータ幅を超えた場合でも、可能なだけ結果を返すようにしたい。
 * ***************************************/

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Signal_Process
{
    /// <summary>
    /// FFT/IFFT関連クラス
    /// 
    ///     開発者：Katsuhiro Morishita(2011.2) Kumamoto-University
    ///     
    ///     更新履歴：2010/5/23   森下整備開始
    /// 
    ///     参考文献:ニューメリカル・レシピ・イン・シー
    ///             用語はこの書籍に準ずる。
    /// 
    /// 
    /// 使用する際には、プロジェクトにファイルを追加した後に、ソースヘッダに以下のコードを追加して下さい。
    /// 「
    /// using Signal_Process;
    /// 」
    /// </summary>
    class FFT
    {
        private int FFT_SIZE;
        private double[] buff;
        public double[] F_re;
        public double[] F_im;

        public enum FFT_point :int{ // データサイズ(2^n)をこの列挙体を使って宣言すれば間違えない.
            size32 = 32, 
            size64 = 64,
            size128 = 128,
            size256 = 256,
            size512 = 512,
            size1024 = 1024,
            size2048 = 2048,
            size4096 = 4096,
            size8192 = 8192
        }
        private enum FFTorIFFT : int { 
            FFTconvert = 1,
            IFFTconvert = -1
        }

        /********************************************
        * 窓関数　ハミング窓（Hanning window）
        *   両端で完全には0にならない(おそらく数値計算上の問題)
        * 
        * ******************************************/
        private double[] Hanning_window(double[] data)
        {
            double wj;
            int j;

            for( j = 0 ; j < data.Length ; j++ )
            {
                wj = 0.5 * (1.0 - Math.Cos(2.0 * Math.PI * (double)j / ((double)data.Length - 1))); // 窓の計数を計算
                data[j] *= wj;
            }
            return data;
        }
        /********************************************
        * 窓関数　Parzen窓(Parzen window)三角窓・・・ハミング窓より周波数のブレが少ないが、サイドローブはやや大きい
        *   引数：データ配列
        *   返値：ウィンドウをかけた後のデータ配列
        * ******************************************/
        private double[] Parzen_window(double[] data) {
            double wj;
            int j;

            for( j = 0 ; j < data.Length ; j++ ) {
                wj = 1.0 - Math.Abs(((double)j - 0.5 * ((double)data.Length - 1)) / (0.5 * ((double)data.Length + 1))); // 窓の計数を計算
                data[j] *= wj;
            }
            return data;
        }
        /********************************************
         * 渡されたデータを記憶する
         *  実部のみのデータを解析するのに使用する。
         *  渡されたデータは自動的に窓関数をかける。
         *  データサイズのチェックも行う。
         *  OKであればtrueを返す。
         *  また、データサイズよりFFTサイズの方が大きい場合は、0で埋める。
         * ******************************************/
        public Boolean Dataset(double[] data , int fft_size) {
            int i,k;
            Boolean n_test = false;

            for( i = 0 ; i < 15 ; i++ ) {               // 処理サイズのチェック
                if( fft_size == 2 << i ) n_test = true; // 2の整数乗でなければならない
            }

            if( n_test ){
                FFT_SIZE = fft_size;
            }else{
                return false;                           // 2の整数乗でないのでエラーとして戻す
            }
            if (data.Length > fft_size) return false;   // 渡されたデータサイズがFFTデータサイズを超えているようであればエラーを返す。
            data = Hanning_window(data);                // データウィンドウをかける
            buff = new double[FFT_SIZE * 2 + 1];        // データサイズを調整
            for( i = 0 ; i < FFT_SIZE ; i++ )           // 受け渡されたデータをコピー
            {
                k = 2 * i + 1;
                if( data.Length > i ){                  // ある限りコピー
                    buff[k] = data[i];
                }else {
                    buff[k] = 0.0;                      // 余ったら0を詰める
                }
                buff[k + 1] = 0.0;
            }
            return true;
        } 
        /**************************************************
         * FFT/IFFT関数
         * 引数：   data[]　非変換変数。結果も格納。長さnnの複素配列。
         *                  配列の中には、実部・虚部・実部・虚部・実部・虚部・実部・虚部・…と並んでおればよい。原本では要素番号1から詰めていることを仮定している。
         *                  実部と虚部が入れ替わっても特に影響は無いはず。，
         *          nn　data[]に格納されたデータ長を表す。要素数の半分である。2の整数乗である必要がある。，
         *          isign　1：FFT変換を行う　-1：IFFT変換を行う。結果はnn倍される。
         ****************************************************/
        public void Calc(double[] data, long nn , int isign ){
            long n, mmax, m, j, istep, i;
            double wtemp, wr, wpr, wpi, wi, theta;              // 三角関数の漸化式用
            double tempr, tempi;

            n = nn << 1;                                        // n:データポインタ数の2倍
            j = 1;
            for (i = 1; i < n; i += 2) {                        // ビット反転アルゴリズム
                if (j > i) {                                    // 複素数を交換
                    tempr = data[j]; data[j] = data[i]; data[i] = tempr;                    // 原本では実部を表す
                    tempr = data[j + 1]; data[j + 1] = data[i + 1]; data[i + 1] = tempr;    // 原本では虚部を表す
                }
                m = n >> 1;
                while (m >= 2 && j > m) {
                    j -= m;
                    m >>= 1;
                }
                j += m;
            }
            mmax = 2;                                           // 以下はDanielson-Lanczosアルゴリズムを採用
            while (n > mmax) {                                  // 外側のループはlog2 nn回実行される
                istep = mmax << 1;
                theta = isign * (6.28318530717959 / mmax);      // 三角関数の漸化式の初期値. 2*pi = 6.28…
                wtemp = Math.Sin (0.5 * theta);
                wpr = -2.0 * wtemp * wtemp;
                wpi = Math.Sin(theta);
                wr = 1.0;
                wi = 0.0;
                for (m = 1; m < mmax; m += 2) {
                    for (i = m; i <= n; i += istep) {
                        j = i + mmax;                           // Danielson-Lanczos公式
                        tempr = wr * data[j] - wi * data[j + 1];// 原本通りならreal part
                        tempi = wr * data[j + 1] + wi * data[j];// 原本通りならimaginary part
                        data[j] = data[i] - tempr;
                        data[j + 1] = data[i + 1] - tempi;
                        data[i] += tempr;
                        data[i+1] += tempi;
                    }                                           // 三角関数の漸化式
                    wr = (wtemp = wr) * wpr - wi * wpi + wr;    // ここは処理をひっくり返せばwtempをそのままwrに置き換えられるが、恐らくrealパートから計算したいのだろう。
                    wi = wi * wpr + wtemp * wpi + wi;
                }
                mmax = istep;
            }
            return;
        }
        /********************************************
         * 実部のみのデータを解析する
         *  鳥の声を解析するときはこれを呼び出せばよい。
         *  例：
         *      fft a = new fft();                              // このクラスを宣言して、実態も確保
         *      a.fft_culc(temp , (int)fft.fft_point.size256);  // 実数配列を渡してFFT演算を実行。配列サイズがFFTデータ幅と一致するように注意。
         *  
         *  FFT結果はF_reとF_imに格納される。解析後に外部からアクセスると良い。
         *  結果の解析も連続的に行う場合は次のように呼び出す。
         *  
         *  if(a.fft_culc(temp , (int)fft.fft_point.size256)){
         *      何か処理;
         *  }
         * 
         *  引数：　data：データ配列，size：FFTデータ幅
         *  返値：成功するとtrueを返す。
         *  
         * 
         *  将来の拡張予定：もしかすると、このクラスのインスタンスを生成する際に、new()にFFTデータ幅を格納するかもしれない。
         *                  その場合、引数が減る…。が、変調スペクトル専用も宣言することになって何だかやだなぁ。
         * ******************************************/
        public Boolean  FFT_calc(double[] data, int size)
        {
            Boolean ans;
            int i,k;
            double[] data_copy = (double[])data.Clone();    // 受け渡された配列が変化しないようにディープコピー

            ans = Dataset(data_copy, size);                 // データの前処理
            if( ans ) {                                     // エラーがなければ実行する
                Calc(buff,size,(int)FFT.FFTorIFFT.FFTconvert );      // FFT演算
                F_re = new double[size];                    // サイズの決定。
                F_im = new double[size];                    // サイズの決定。

                for( i = 0 ; i < size ; i++ )
                {
                    k = 2 * i + 1;
                    F_re[i] = buff[k];                      // 結果を格納する
                    F_im[i] = buff[k + 1];                  // こっちはimaginaryパート
                }
            }
            return ans;
        }
    }
}
